#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    _Back.py

    The inverse Radon transform with non-iterative techniques.
"""

from __future__ import division

import sys

from . import _Back as back2d
from . import _Back_iterative as back2d_it
from . import _Back_3D as back3d

__all__ = back2d.__all__ + back2d_it.__all__


def art(sinogram, angles, initial=None, iterations=1,
        jmc=None, jmm=None):
    u""" Algebraic Reconstruction Technique

    The Algebraic Reconstruction Technique (ART) iteratively
    computes the inverse of the Radon transform in two dimensions.
    The reconstruction technique uses *rays* of the diameter of
    one pixel to iteratively solve the system of linear equations
    that describe the projection process. The binary weighting
    factors are

     - 1, if the center of the a pixel is within the *ray*
     - 0, else


    Parameters
    ----------
    sinogram : ndarray, shape (A,N) or (A,M,N)
        Two- or three-dimensional sinogram of line recordings. If the
        sinogram is three-dimensional, the dimension `M` iterates 
        through the slices. The rotation takes place through
        the second (y) axis.
    angles : ndarray, length A
        Angular positions of the `sinogram` in radians. The angles
        at which the sinogram slices were recorded do not have to be
        distributed equidistantly as in :func:`backproject`.
        The angles are internaly converted to modulo PI.
    initial : ndarray, shape (N,N) or (N,M,N), optional
        The initial guess for the solution.
    iterations : int
        Number of iterations to perform.
    jmc, jmm : instance of :func:`multiprocessing.Value` or `None`
        The progress of this function can be monitored with the 
        :mod:`jobmanager` package. The current step `jmc.value` is
        incremented `jmm.value` times. `jmm.value` is set at the 
        beginning.


    Returns
    -------
    out : ndarray, shape (N,N) or (N,M,N)
        The reconstructed image.


    See Also
    --------
    sart : simultaneous algebraic reconstruction technique


    Notes
    -----
    For theoretical backround, see
    Kak, A. C., & Slaney, M.. *Principles of Computerized
    Tomographic Imaging*, SIAM, (2001)

    Sec. 7.2:
    *"ART reconstrutions usually suffer from salt and pepper noise,
    which is caused by the inconsitencies introuced in the set of
    equations by the approximations commonly used for* 
    :math:`w_{ik}` *'s."*
    """
    # This is a convenience function that wraps 2d and 3d
    # collect all arguments
    code = sys._getframe().f_code
    keys = code.co_varnames[:code.co_argcount]
    kwargs = dict()
    lcs = locals()
    for key in list(keys):
        kwargs[key] = lcs[key]

    return two_three_dim_recon(code, kwargs)


def backproject(sinogram, angles, filtering="ramp",
                padding=True, padval=0,
                jmc=None, jmm=None, verbose=0):
    u""" Backprojection with the Fourier slice theorem

    Computes the inverse of the radon transform using filtered
    backprojection.


    Parameters
    ----------
    sinogram : ndarray, shape (A,N) or (A,M,N)
        Two- or three-dimensional sinogram of line recordings. If the
        sinogram is three-dimensional, the dimension `M` iterates 
        through the slices. The rotation takes place through
        the second (y) axis.
    angles : (A,) ndarray
        Angular positions of the `sinogram` in radians equally
        distributed from zero to PI.
    filtering : {'ramp', 'shepp-logan', 'cosine', 'hamming', \
                 'hann'}, optional
        Specifies the Fourier filter. Either of

        ``ramp``
          mathematically correct reconstruction

        ``shepp-logan``

        ``cosine``

        ``hamming``

        ``hann``

    padding : bool, optional
        Pad the input data to the second next power of 2 before
        Fourier transforming. This reduces artifacts and speeds up
        the process for input image sizes that are not powers of 2.
    padval : float
        The value used for padding.
        If `padval` is `None`, then the edge values are used for
        padding (see documentation of `numpy.pad`).
    jmc, jmm : instance of :func:`multiprocessing.Value` or `None`
        The progress of this function can be monitored with the 
        :mod:`jobmanager` package. The current step `jmc.value` is
        incremented `jmm.value` times. `jmm.value` is set at the 
        beginning.
    verbose : int
        Increment to increase verbosity.


    Returns
    -------
    out : ndarray, shape (N,N) or (N,M,N)
        The reconstructed image.


    Examples
    --------
    >>> import numpy as np
    >>> from radontea import backproject
    >>> N = 5
    >>> A = 7
    >>> x = np.linspace(-N/2, N/2, N)
    >>> projection = np.exp(-x**2)
    >>> sinogram = np.tile(projection, A).reshape(A,N)
    >>> angles = np.linspace(0, np.pi, A, endpoint=False)
    >>> backproject(sinogram, angles)
    array([[ 0.03768159, -0.02020189, -0.02287825,  0.03766525,  0.01771381],
           [ 0.03945197, -0.01662539,  0.02187192,  0.04124174,  0.0194842 ],
           [ 0.02491933,  0.11287706,  0.41412685,  0.17074419,  0.00495156],
           [ 0.03915731,  0.10966977,  0.27540441,  0.16753691,  0.01918954],
           [ 0.02052438,  0.03156967,  0.07341154,  0.08943681,  0.0005566 ]])


    See Also
    --------
    fourier_map : implementation by Fourier interpolation
    sum : implementation by summation in real space
    """
    # This is a convenience function that wraps 2d and 3d
    # collect all arguments
    code = sys._getframe().f_code
    keys = code.co_varnames[:code.co_argcount]
    kwargs = dict()
    lcs = locals()
    for key in list(keys):
        kwargs[key] = lcs[key]

    return two_three_dim_recon(code, kwargs)


def fourier_map(sinogram, angles, intp_method="cubic",
                jmc=None, jmm=None):
    u""" Fourier mapping with the Fourier slice theorem

    Computes the inverse of the radon transform using Fourier
    interpolation.
    Warning: This is the naive reconstruction that assumes that
    the image is rotated through the upper left pixel nearest to
    the actual center of the image. We do not have this problem for
    odd images, only for even images.


    Parameters
    ----------
    sinogram : ndarray, shape (A,N) or (A,N,M)
        Two- or three-dimensional sinogram of line recordings. If the
        sinogram is three-dimensional, the dimension `M` iterates 
        through the slices. The rotation takes place through
        the second (y) axis.
    angles : (A,) ndarray
        Angular positions of the `sinogram` in radians equally
        distributed from zero to PI.
    intp_method : {'cubic', 'nearest', 'linear'}, optional
        Method of interpolation. For more information see
        `scipy.interpolate.griddata`. One of

        ``nearest``
          instead of interpolating, use the points closest to
          the input data.

        ``linear``
          bilinear interpolation between data points.

        ``cubic``
          interpolate using a two-dimensional poolynimial surface.

    jmc, jmm : instance of :func:`multiprocessing.Value` or `None`
        The progress of this function can be monitored with the 
        :mod:`jobmanager` package. The current step `jmc.value` is
        incremented `jmm.value` times. `jmm.value` is set at the 
        beginning.


    Returns
    -------
    out : ndarray, shape (N,N) or (N,M,N)
        The reconstructed image.


    See Also
    --------
    backproject : implementation by backprojection
    sum : implementation by summation in real space
    scipy.interpolate.griddata : the used interpolation method
    """
    # This is a convenience function that wraps 2d and 3d
    # collect all arguments
    code = sys._getframe().f_code
    keys = code.co_varnames[:code.co_argcount]
    kwargs = dict()
    lcs = locals()
    for key in list(keys):
        kwargs[key] = lcs[key]

    return two_three_dim_recon(code, kwargs)


def sart(sinogram, angles, initial=None, iterations=1,
         jmc=None, jmm=None):
    u""" Simultaneous Algebraic Reconstruction Technique


    SART computes an inverse of the Radon transform in two dimensions.
    The reconstruction technique uses "rays" of the diameter of
    one pixel to iteratively solve the system of linear equations
    that describe the image. The weighting factors are bilinear
    elements. At the beginning and end of each ray, only partial
    weights are used. The pixel values of the image are updated
    only after each iteration is complete.


    Parameters
    ----------
    sinogram : ndarray, shape (A,N) or (A,M,N)
        Two- or three-dimensional sinogram of line recordings. If the
        sinogram is three-dimensional, the dimension `M` iterates 
        through the slices. The rotation takes place through
        the second (y) axis.
    angles : ndarray, length A
        Angular positions of the `sinogram` in radians. The angles
        at which the sinogram slices were recorded do not have to be
        distributed equidistantly as in backprojection techniques.
        The angles are internaly converted to modulo PI.
    initial : ndarray, shape (N,N) or (N,M,N), optional
        The initial guess for the solution.
    iterations : integer
        Number of iterations to perform.
    jmc, jmm : instance of :func:`multiprocessing.Value` or `None`
        The progress of this function can be monitored with the 
        :mod:`jobmanager` package. The current step `jmc.value` is
        incremented `jmm.value` times. `jmm.value` is set at the 
        beginning.


    Returns
    -------
    out : ndarray, shape (N,N) or (N,M,N)
        The reconstructed image.


    See Also
    --------
    art : algebraic reconstruction technique


    Notes
    -----
    Algebraic reconstruction technique (ART) (see `art`):
        Iterations are performed over each ray of each projection.
        Weighting factors are binary (1 if center of pixel is
        within ray, 0 else). This leads to salt and pepper noise.

    Simultaneous iterative reconstruction technique (SIRT):
        Same idea as ART, but for each iteration, the change of the
        image f is computed for all rays and projections separately
        and the weights are applied simultaneously after each
        iteration. The result is a slower convergence but the final
        image is also less noisy.

    This implementation does NOT use a hamming window to filter
    the data and to emphasize points at the center of the recon-
    struction region.

    For theoretical backround, see
    Kak, A. C., & Slaney, M.. *Principles of Computerized
    Tomographic Imaging*, SIAM, (2001)

    Sec 7.4:
    *"[SART] seems to combine the best of ART and SIRT. [...] Here
    are the main features if SART: First, [...] the traditional
    pixel basis is abandonded in favor of bilinear elements
    [e.g. interpolation]. Also, for a circular reconstruction
    region, only partial weights are assigned to the first and last
    picture elements on the individual rays. To further reduce the
    noise [...], the correction terms are simultaneously applied
    for all the rays in one projection [...]."*
    """
    # This is a convenience function that wraps 2d and 3d
    # collect all arguments
    code = sys._getframe().f_code
    keys = code.co_varnames[:code.co_argcount]
    kwargs = dict()
    lcs = locals()
    for key in list(keys):
        kwargs[key] = lcs[key]

    return two_three_dim_recon(code, kwargs)


def sum(sinogram, angles, jmc=None, jmm=None):  # @ReservedAssignment
    u""" Sum-reconstruction with the Fourier slice theorem

    Computes the inverse of the radon transform by computing the
    integral in real space.

    .. warning:: This algorithm is slow and prone to numerical errors.
    
    Parameters
    ----------
    sinogram : ndarray, shape (A,N) or (A,M,N)
        Two- or three-dimensional sinogram of line recordings. If the
        sinogram is three-dimensional, the dimension `M` iterates 
        through the slices. The rotation takes place through
        the second (y) axis.
    angles : (A,) ndarray
        Angular positions of the `sinogram` in radians equally 
        distributed from zero to PI.
    jmc, jmm : instance of :func:`multiprocessing.Value` or `None`
        The progress of this function can be monitored with the 
        :mod:`jobmanager` package. The current step `jmc.value` is
        incremented `jmm.value` times. `jmm.value` is set at the 
        beginning.



    Returns
    -------
    out : ndarray, shape (N,N) or (N,M,N)
        The reconstructed image.


    See Also
    --------
    backproject : implementation by backprojection
    fourier_map : implementation by summation in real space
    """
    # This is a convenience function that wraps 2d and 3d
    # collect all arguments
    code = sys._getframe().f_code
    keys = code.co_varnames[:code.co_argcount]
    kwargs = dict()
    lcs = locals()
    for key in list(keys):
        kwargs[key] = lcs[key]

    return two_three_dim_recon(code, kwargs)


def two_three_dim_recon(code, kwargs):
    sinogram = kwargs["sinogram"]

    if len(sinogram.shape) == 2:
        # 2D
        if hasattr(back2d, code.co_name):
            func = getattr(back2d, code.co_name)
        elif hasattr(back2d_it, code.co_name):
            func = getattr(back2d_it, code.co_name)
        else:
            raise NotImplementedError("Unknown method: ", code.co_name)
    elif len(sinogram.shape) == 3:
        # 3D
        if hasattr(back3d, code.co_name):
            func = getattr(back3d, code.co_name)
        else:
            raise NotImplementedError("Unknown method: ", code.co_name)
    else:
        raise ValueError("sinogram must have dimension 2 or 3.")

    return func(**kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
u"""

There are several methods to compute the inverse **Radon** transform.
The module :py:mod:`radontea` implements some of them. I focussed on
code readability and thorough comments. The result is a collection of
algorithms that are suitable for **tea**\ching the basics of
computerized tomography.


Recommended literature 
----------------------
- Aninash C. Kak and Malcom Slaney. *Principles of Computerized 
  Tomographic Imaging*. Ed. by Robert E. O’Malley. SIAM, 2001, p. 327.
  ISBM: 089871494X.

- Johann Radon. *Über die Bestimmung von Funktionen durch ihre
  Integralwerte längs gewisser Mannigfaltigkeiten*. Tech. rep. Leipzig:
  Berichte über die Verhandlungen der Königlich-Sächsischen Gesellschaft
  der Wissenschaften zu Leipzig, 1917, pp. 262–277.

- R A Crowther, D J DeRosier, and A Klug. *The Reconstruction of a
  Three-Dimensional Structure from Projections and its Application to
  Electron Microscopy*. In: Proceedings of the Royal Society of London.
  A. Mathematical and Physical Sciences 317.1530 (1970), pp. 319–340.
  doi: `10.1098/rspa.1970.0119 
  <http://dx.doi.org/10.1098/rspa.1970.0119>`_.

Obtaining radontea 
------------------
If you have Python and :py:mod:`numpy` installed, simply run

    pip install radontea

The source code of radontea is available at
https://github.com/RI-imaging/radontea.


Citing radontea
---------------
Please cite this package if you are using it in a scientific
publication.

This package should be cited like this 
(replace "x.x.x" with the actual version of radontea that you used):

.. topic:: cite

    Paul Müller (2013) *radontea: Python algorithms for the inversion of
    the Radon transform* (Version x.x.x) [Software].
    Available at https://pypi.python.org/pypi/radontea/


You can find out what version you are using by typing
(in a Python console):


    >>> import radontea
    >>> radontea.__version__
    '0.1.4'

"""

from ._Back_meta import *
#from . import _Back as _2d
#from . import _Back_3D as _3d
from ._Back_Fan import *
#from ._Back_iterative import *
from ._Radon import *

from ._version import version as __version__
__author__ = u"Paul Müller"
__license__ = "BSD (3-Clause)"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Tests backprojection algorithm
"""
from __future__ import division, print_function

import numpy as np
import os
from os.path import join
import sys
import time

from os.path import abspath, dirname, split

# Add parent directory to beginning of path variable
DIR = dirname(abspath(__file__))
sys.path = [split(DIR)[0]] + sys.path

import radontea
import radontea._Back


def create_test_sino(A, N):
    """
    Creates test sinogram.
    """
    resar = np.zeros((A,N), dtype=float)
    angles = np.linspace(0, np.pi, A, endpoint=False)
    x = np.linspace(-N/2, N/2, N, endpoint=True)
    dev = np.sqrt(N/2)
    off = N/7
    for ii in range(A):
        # Gaussian distribution sinogram
        x0 = np.cos(angles[ii])*off
        y = np.exp(-(x-x0)**2/dev**2)
        resar[ii] = y

    # some normalization, so that max of Radon-inversion
    # is about 1.    
    resar = 2*dev*resar/resar.max()
    return resar, angles


def test_2d_backproject():
    myname = sys._getframe().f_code.co_name
    print("running ", myname)
    
    sino, angles = create_test_sino(A=100, N=100)
    r = radontea.backproject(sino, angles, padding=True)
    r2 = radontea._Back.backproject(sino, angles, padding=True)
    
    #np.savetxt('outfile.txt', np.array(r).flatten().view(float), fmt="%.8f")
    assert np.allclose(r, r2)
    assert np.allclose(np.array(r).flatten().view(float), results[myname])



# Get results
results = dict()
datadir = join(DIR, "data")
for f in os.listdir(datadir):
    #np.savetxt('outfile.txt', np.array(r).flatten().view(float), fmt="%.10f")
    #np.savetxt('outfile.txt', np.array(r).flatten().view(float))
    glob = globals()
    if f.endswith(".txt") and f[:-4] in list(glob.keys()):
        results[f[:-4]] = np.loadtxt(join(datadir, f))


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Example of handling 3D data.

Using the `*_3d` methods only speeds up calculations if your processor
has multiple cores.
"""
from __future__ import print_function, division
from multiprocessing import cpu_count
import numpy as np
from os.path import split, dirname, abspath
import time
import sys

sys.path = [split(dirname(abspath(__file__)))[0]] + sys.path

import radontea as rt


try:
    # use jobmanager if available
    import jobmanager as jm
    jm.decorators.decorate_module_ProgressBar(rt, 
                        decorator=jm.decorators.ProgressBarOverrideCount,
                        interval=.1)
except:
    pass


if __name__ == "__main__":
    A = 70    # number of angles
    N = 128   # detector size x
    M = 24    # detector size y (number of slices)
    
    # generate random data
    sino0 = np.random.random((A,N))     # for 2d example
    sino = np.random.random((A,M,N))    # for 3d example
    sino[:,0,:] = sino0
    angles = np.linspace(0,np.pi,A)     # for both
    
    a = time.time()
    data0 = rt.backproject(sino0, angles)
    print("time on 1 core: {} s".format((time.time() - a)*M))
    
    a = time.time()
    data = rt._Back_3D.backproject(sino, angles)
    print("time on {} cores: {} s".format(cpu_count(), time.time() - a))
    
    
    assert np.sum(data0==data[:,0,:]) == N**2, "2D and 3D results don't match"
    



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# To create a distribution package for pip or easy-install:
# python setup.py sdist
from setuptools import setup
from os.path import exists, dirname, realpath
import sys


author = u"Paul Müller"
authors = [author]
description = 'Collection of algorithms to compute the inverse Radon transform'
name = 'radontea'
year = "2014"

sys.path.insert(0, realpath(dirname(__file__))+"/"+name)
try:
    from _version import version
except:
    version = "unknown"


if __name__ == "__main__":
    setup(
        name=name,
        author=author,
        author_email="paul.mueller@biotec.tu-dresden.de",
        version=version,
        license="BSD (3 clause)",
        url='https://github.com/RI-imaging/radontea',
        packages=[name],
        package_dir={name: name},
        description=description,
        long_description=open('README.rst').read() if exists('README.rst') else '',
        install_requires=[ "NumPy >= 1.5.1", "SciPy >= 0.8.0"],
        keywords=["tomography", "ct", "radon", "computerized tomography",
                  "optical projection tomography"],
        setup_requires=['pytest-runner'],
        tests_require=["pytest"],
        classifiers= [
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.4',
            'Topic :: Scientific/Engineering :: Visualization',
            'Intended Audience :: Science/Research'
                     ],
        platforms=['ALL'],
        )




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    _logo.py
    
    Functions that are used to create the logo of radontea.
"""
from __future__ import division

import numpy as np

if __name__ == "__main__":
    from _Radon import radon  # @UnusedImport
else:
    from ._Radon import radon  # @Reimport


def logo(x, y, N):
    """
        Vector representation of radonteach logo.


        Parameters
        ----------
        x,y : ndarray or float
            coordinates where to calculate the logo.
        N : float
            total size of the image (NxN)


        Returns
        -------
        v : ndarray
            inverted values of the logo at the coordinates (x,y)
            normalized to one.
    """
    z1 = np.exp(-((x - N / 8)**2 + (y + N / 14)**2) / (N / 8)**2)
    z2 = np.exp(-((x + N / 10)**2 + (y - N / 10)**2) / (N / 8)**2)

    z3 = np.exp(-((x - N / 8)**2 + (y + N / 10)**2) / (N / 12)**2)
    z4 = np.exp(-((x + N / 10)**2 + (y - N / 14)**2) / (N / 12)**2)

    z5 = np.exp(-((x)**2 + (y)**2) / (N / 4)**2)
    z = (z1 + z2 - (z3 + z4))
    z = (z - z.min()) * z5
    z /= z.max()
    z[np.where(x**2 + y**2 >= (N / 2)**2)] = 0

    return z


def get_original(N=64):
    """
        Returns a two-dimensional square image that is the image used
        to create the sinogram for the radontea logo.
    """
    x = np.linspace(-N / 2, N / 2, N, endpoint=False)
    X = x.reshape(1, -1)
    Y = x.reshape(-1, 1)
    z = logo(X, Y, N)

    return np.array((z) * 255, dtype=np.uint16)


def get_logo(N=64):
    """
        Returns a two-dimensional square logo of resolution NxN. This
        function discretizes the vector image representation of the
        radonteach logo.
    """
    a = (get_original(N=N))
    angles = np.linspace(0, np.pi * 1.6, N)
    sinogram = radon(a, angles)
    sinogram = (1 - sinogram / sinogram.max()) * 255
    return sinogram.transpose()


def main():
    # Show the logo
    from matplotlib import pylab as plt
    from matplotlib import cm
    logo = get_logo(N=128)
    plt.imshow(logo, cmap=cm.gray)  # @UndefinedVariable
    plt.axis("off")
    plt.tight_layout()
    plt.show()


if __name__ == "__main__":
    main()
